main(N) :- load_files('arguments.txt'), arguments(Rules1,Rules2,Axioms,Obs),
  apply_all_rules(Rules2,Axioms,Obs,Concls), append(Axioms,Concls,Beliefs), nl,nl,liner(Beliefs),nl,statistics,nl,nl,
%  apply_all_rules(Rules1,Beliefs,Obs,_).
  main2(N,Rules1,Rules2,Beliefs,Obs).

main2(0,_,_,_,_) :- !.
main2(N,Rules1,Rules2,Beliefs1,Obs) :-
  apply_all_rules(Rules1,Beliefs1,Obs,Beliefs2), apply_all_rules(Rules2,Beliefs2,Obs,Beliefs3), 
  append(Beliefs2,Beliefs3,Beliefs4), nl,nl,liner(Beliefs4),nl,statistics,nl,nl,
  N1 is N-1, main2(N1,Rules1,Rules2,Beliefs4,Obs).

liner([X|Xs]) :- print(X), nl, open(result,append,OUTSTR), write(OUTSTR,X),nl(OUTSTR),close(OUTSTR), liner(Xs).
liner([]).

% ------------------------------------------------------------------------------------
% Applying all the inference rules to a belief set
% ------------------------------------------------------------------------------------

apply_all_rules([],_,_,[]).
apply_all_rules([Rule|Rules],Beliefs1,Obs,Concls4) :-
  uni_belief_set(Beliefs1,Rule,Beliefs2,N),
  all_groups(Beliefs2,Beliefs2,Obs,N,Rule,Group),
  repl_all_concl(Group,Concls1), del_doubles(Concls1,Concls2),
  apply_all_rules(Rules,Beliefs1,Obs,Concls3),
  append(Concls2,Concls3,Concls4).

% ------------------------------------------------------------------------------------
% Finding all the groups of beliefs validing the premises of an inference rule
% ------------------------------------------------------------------------------------

all_groups([],_,_,_,_,[]).
all_groups([Belief|Beliefs],AllBeliefs,Obs,N1,rule([Prem|Prems1],Concl1),Group3) :- is_in(all(in,Name,X),Prem),
  belief_flex_corr(Name,Prem,Belief,L), !,
  repl_all_flex(all(in,Name,X),N1,L,[],rule([Prem|Prems1],Concl1),rule(Prems2,Concl2),N2),
  all_groups(AllBeliefs,AllBeliefs,Obs,N2,rule(Prems2,Concl2),Group1),
  all_groups(Beliefs,AllBeliefs,Obs,N1,rule([Prem|Prems1],Concl1),Group2),
  append(Group1,Group2,Group3).
all_groups([_|Beliefs],AllBeliefs,Obs,N1,rule([Prem|Prems1],Concl),Group) :- is_in(all(in,_,_),Prem), !,
  all_groups(Beliefs,AllBeliefs,Obs,N1,rule([Prem|Prems1],Concl),Group).
all_groups(Beliefs,AllBeliefs,Obs,N,rule([Prem|Prems1],Concl),Group) :- is_in(all(out,_,_),Prem), !,
  append(Prems1,[Prem],Prems2), all_groups(Beliefs,AllBeliefs,Obs,N,rule(Prems2,Concl),Group).

all_groups(AllBeliefs,AllBeliefs,Obs1,N,rule([[spec(obs),Prem]|Prems1],Concl1),Group4) :-
  select(Belief,Obs1,Obs2), multi_correl(Prem,Belief,Vars1), clean_vars(Vars1,Vars2), !,
  rule_replace(rule(Prems1,Concl1),Vars2,rule(Prems2,Concl2)),
  all_groups(AllBeliefs,AllBeliefs,Obs2,N,rule(Prems2,Concl2),Group1), map_prem(Belief,Group1,Group2),
  all_groups(AllBeliefs,AllBeliefs,Obs2,N,rule([[spec(obs),Prem]|Prems1],Concl1),Group3),
  append(Group2,Group3,Group4).
all_groups(_,_,_,_,rule([[spec(obs),_]|_],_),[]) :- !.

/*
all_groups(Allbeliefs,Allbeliefs,Obs,N,rule([[or,Cond|_]|Prems],Concl),Group) :-
  all_groups(Allbeliefs,Allbeliefs,Obs,N,rule([Cond|Prems],Concl),Group), Group \= [], !.
all_groups(Allbeliefs,Allbeliefs,Obs,N,rule([[or,_|Conds]|Prems],Concl),Group) :-
  all_groups(Allbeliefs,Allbeliefs,Obs,N,rule([[or|Conds]|Prems],Concl),Group), !.
all_groups(Allbeliefs,Allbeliefs,Obs,N,rule([[or,Cond]|Prems],Concl),Group) :-
  all_groups(Allbeliefs,Allbeliefs,Obs,N,rule([Cond|Prems],Concl),Group), !.
*/
all_groups(Allbeliefs,Allbeliefs,Obs,N,rule([[spec(not,Cond),Prem]|Prems],Concl),Group) :-
  member(Cond,[r_csf,f_forms,=,<]), \+ condition(Cond,Allbeliefs,Prem), !,
  all_groups(Allbeliefs,Allbeliefs,Obs,N,rule(Prems,Concl),Group).
all_groups(Allbeliefs,Allbeliefs,Obs,N,rule([[spec(Cond),Prem]|Prems],Concl),Group) :-
  member(Cond,[r_csf,f_forms,=,<]), condition(Cond,Allbeliefs,Prem), !,
  all_groups(Allbeliefs,Allbeliefs,Obs,N,rule(Prems,Concl),Group).
all_groups(_,_,_,_,rule([[spec(not,Cond),_]|_],_),[]) :-
  member(Cond,[r_csf,f_forms,=,<]), !.
all_groups(_,_,_,_,rule([[spec(Cond),_]|_],_),[]) :-
  member(Cond,[r_csf,f_forms,=,<]), !.

all_groups(AllBeliefs,AllBeliefs,Obs,N,rule([[spec(f_csf),Model,Prem]|Prems],Concl),Group) :- !,
  belief_filter(AllBeliefs,Model,ReducedBeliefs), all_csf(ReducedBeliefs,AllCSFs),
  all_groups(AllCSFs,AllBeliefs,Obs,N,rule([[spec(f_csf),Model,Prem]|Prems],Concl),Group).
all_groups(AllCSFs1,AllBeliefs,Obs,N,rule([[spec(f_csf),Model,Prem]|Prems1],Concl1),Group4) :-
  select(Csf,AllCSFs1,AllCSFs2), correlation(Prem,Csf,Vars1), clean_vars(Vars1,Vars2), !,
  rule_replace(rule(Prems1,Concl1),Vars2,rule(Prems2,Concl2)),
  all_groups(AllBeliefs,AllBeliefs,Obs,N,rule(Prems2,Concl2),Group1), map_prem([fcsf,Model,Csf],Group1,Group2),
  all_groups(AllCSFs2,AllBeliefs,Obs,N,rule([[spec(f_csf),Model,Prem]|Prems1],Concl1),Group3),
  append(Group2,Group3,Group4).

all_groups(_,_,_,_,rule([],Concl),[rule([],Concl)]) :- !.
all_groups([Belief|Beliefs],AllBeliefs1,Obs,N1,rule([Prem|Prems1],Concl1),Group4) :-
  multi_correl(Prem,Belief,Vars1), clean_vars(Vars1,Vars2), !,
  rule_replace(rule(Prems1,Concl1),Vars2,rule(Prems2,Concl2)),
  upgrade_vars(N1,Belief,AllBeliefs1,AllBeliefs2,N2),
  all_groups(AllBeliefs2,AllBeliefs2,Obs,N2,rule(Prems2,Concl2),Group1), map_prem(Belief,Group1,Group2),
  all_groups(Beliefs,AllBeliefs1,Obs,N1,rule([Prem|Prems1],Concl1),Group3),
  append(Group2,Group3,Group4).
all_groups([_|Beliefs],AllBeliefs,Obs,N,rule([Prem|Prems1],Concl1),Group) :-  
  all_groups(Beliefs,AllBeliefs,Obs,N,rule([Prem|Prems1],Concl1),Group).

% ------------------------------------------------------------------------------------
% Analyzing and modifying the content of formulae
% ------------------------------------------------------------------------------------

% Correlation between a formula of an inference rule and a formula of the belief set
correlation(X,X,[]) :- !.
correlation(var(X),Y,[corr(var(X),Y)]) :- !.
correlation(Y,var(X),[corr(var(X),Y)]) :- !.
correlation([X|Xs],[Y|Ys],Ls) :- correlation(X,Y,Z), correlation(Xs,Ys,Zs), append(Z,Zs,Ls).
correlation(pred(X1,Y1),pred(X2,Y2),Z3) :- correlation(X1,X2,Z1), correlation(Y1,Y2,Z2), append(Z1,Z2,Z3).
correlation([Fct|Xs1],Y,[corr(var(X),Xs2)]) :- member(Fct,[+,-]), is_in(var(X),Xs1), !, invert_op(var(X),[Fct|Xs1],Y,Xs2).
correlation(Y,[Fct|Xs1],[corr(var(X),Xs2)]) :- member(Fct,[+,-]), is_in(var(X),Xs1), !, invert_op(var(X),[Fct|Xs1],Y,Xs2).
% Correlation between a premise of an inference rule and a belief
multi_correl([],[],[]).
multi_correl([X|Xs],[Y|Ys],Ls) :- correlation(X,Y,Z), multi_correl(Xs,Ys,Zs), append(Z,Zs,Ls).

% Replacement in a formula of all the instances of certain variables
replace(X,Vars,Y) :- member(corr(X,Y),Vars), !.
replace([+|Xs1],Vars,X) :- replace(Xs1,Vars,Xs2), \+ not_numbers(Xs2), !, addition(Xs2,X).
replace([+|Xs1],Vars,[+|Xs2]) :- !, replace(Xs1,Vars,Xs2).
replace([-|Xs1],Vars,X) :- replace(Xs1,Vars,Xs2), \+ not_numbers(Xs2), !, substraction(Xs2,X).
replace([-|Xs1],Vars,[-|Xs2]) :- !, replace(Xs1,Vars,Xs2).
replace([max|Xs1],Vars,X) :- replace(Xs1,Vars,Xs2), \+ not_numbers(Xs2), !, maximum(Xs2,X).
replace([max|Xs1],Vars,[max|Xs2]) :- !, replace(Xs1,Vars,Xs2).
replace([min|Xs1],Vars,X) :- replace(Xs1,Vars,Xs2), \+ not_numbers(Xs2), !, minimum(Xs2,X).
replace([min|Xs1],Vars,[min|Xs2]) :- !, replace(Xs1,Vars,Xs2).
replace([X|Xs],Vars,[Y|Ys]) :- !, replace(X,Vars,Y), replace(Xs,Vars,Ys).
replace(pred(X1,Y1),Vars,pred(X2,Y2)) :- !, replace(X1,Vars,X2), replace(Y1,Vars,Y2).
replace(all(X,Y,Zs1),Vars,all(X,Y,Zs2)) :- !, replace(Zs1,Vars,Zs2).
replace(X,_,X).
% Replacement in a belief of all the instances of certain variables
multi_replace([],_,[]).
multi_replace([X|Xs],Vars,[Y|Ys]) :- replace(X,Vars,Y), multi_replace(Xs,Vars,Ys).
% Replacement in an entire rule of all the instances of certain variables
rule_replace(rule([],Concl1),Vars,rule([],Concl2)) :- multi_replace(Concl1,Vars,Concl2).
rule_replace(rule([X|Xs],Concl1),Vars,rule([Y|Ys],Concl2)) :-
  multi_replace(X,Vars,Y), rule_replace(rule(Xs,Concl1),Vars,rule(Ys,Concl2)).

% Putting an element as the first premise of each inference rule of a list of inference rules
map_prem(_,[],[]).
map_prem(Prem,[rule(Prems,Concl)|Ls1],[rule([Prem|Prems],Concl)|Ls2]) :- map_prem(Prem,Ls1,Ls2).

addition([X],X).
addition([X|Xs],Y2) :- addition(Xs,Y1), Y2 is Y1+X.
substraction([X],X).
substraction([X|Xs],Y2) :- addition(Xs,Y1), Y2 is X-Y1.
minimum([X],X).
minimum([X|Xs],X) :- minimum(Xs,Y), X<Y, !.
minimum([_|Xs],Y) :- minimum(Xs,Y).
maximum([X],X).
maximum([X|Xs],X) :- maximum(Xs,Y), X>Y, !.
maximum([_|Xs],Y) :- maximum(Xs,Y).

not_numbers(Xs) :- member(X,Xs), \+ number(X).

% ------------------------------------------------------------------------------------
% Replacing all the universal variables (under the form any(X)) in the belief set by
% variables for the inference rule (under the form var(Y))
% ------------------------------------------------------------------------------------

% looking for the maximum number used in the variables of a formula so that we will
% be able to create new variables with different names (numbers actually) than the
% variables already used
var_max([],0) :- !.
var_max(var(X),0) :- \+ number(X), !.
var_max(var(X),X) :- !.
var_max([X|Xs],N1) :- var_max(X,N1), var_max(Xs,N2), N1>N2, !.
var_max([_|Xs],N) :- !, var_max(Xs,N).
var_max(pred(X,Xs),N1) :- var_max(X,N1), var_max(Xs,N2), N1>N2, !.
var_max(pred(_,Xs),N) :- !, var_max(Xs,N).
var_max(all(_,_,X),N) :- !, var_max(X,N).
var_max(_,0).
% looking for the maximum number used in the variables of a premise
multi_var_max([],0).
multi_var_max([X|Xs],N1) :- var_max(X,N1), multi_var_max(Xs,N2), N1>N2, !.
multi_var_max([_|Xs],N) :- multi_var_max(Xs,N).
% looking for the maximum number used in the variables of an entire inference rule
rule_var_max(rule([],Concl),N) :- multi_var_max(Concl,N).
rule_var_max(rule([X|Xs],Concl),N1) :- multi_var_max(X,N1), rule_var_max(rule(Xs,Concl),N2), N1>N2, !.
rule_var_max(rule([_|Xs],Concl),N) :- rule_var_max(rule(Xs,Concl),N).

% looking for all the universal variables in a formula of a belief
uni_vars([],[]) :- !.
uni_vars(any(X),[any(X)]) :- !.
uni_vars([X|Xs],Ls) :- !, uni_vars(X,Y), uni_vars(Xs,Ys), append(Y,Ys,Ls).
uni_vars(pred(X,Xs),Ls) :- !, uni_vars(X,Y), uni_vars(Xs,Ys), append(Y,Ys,Ls).
uni_vars(_,[]).
% looking for all the universal variables in a belief
multi_uni_vars([],[]).
multi_uni_vars([X|Xs],Ls) :- uni_vars(X,Y), multi_uni_vars(Xs,Ys), append(Y,Ys,Ls).

% Deleting the doubles in a list
del_doubles([],[]).
del_doubles([X|Xs],Ys) :- member(X,Xs), !, del_doubles(Xs,Ys).
del_doubles([X|Xs],[X|Ys]) :- del_doubles(Xs,Ys).

% Replacing an universal variable in a formula of a belief
repl_one_uni([],any(_),var(_),[]) :- !.
repl_one_uni(any(X),any(X),var(N),var(N)) :- !.
repl_one_uni([L1|Ls1],any(X),var(N),[L2|Ls2]) :- !, repl_one_uni(L1,any(X),var(N),L2), repl_one_uni(Ls1,any(X),var(N),Ls2).
repl_one_uni(pred(L1,Ls1),any(X),var(N),pred(L2,Ls2)) :- !, repl_one_uni(L1,any(X),var(N),L2), repl_one_uni(Ls1,any(X),var(N),Ls2).
repl_one_uni(X,_,_,X).
% Replacing an universal variable in a belief
multi_repl_one_uni([],any(_),var(_),[]).
multi_repl_one_uni([X|Xs],any(U),var(N),[Y|Ys]) :- repl_one_uni(X,any(U),var(N),Y), multi_repl_one_uni(Xs,any(U),var(N),Ys).
% Replacing all the universal variables in a belief
multi_repl_all_uni(Xs,[],_,Xs).
multi_repl_all_uni(Xs,[any(U)|Uni],N,Zs) :- N1 is N+1, multi_repl_one_uni(Xs,any(U),var(N1),Ys), multi_repl_all_uni(Ys,Uni,N1,Zs).

% Replacing all the universal variables in the belief set
uni_belief_set(Beliefs1,Rule,Beliefs2,N2) :- rule_var_max(Rule,N1), uni_belief_set2(Beliefs1,N1,Beliefs2,N2).
uni_belief_set2([],N,[],N).
uni_belief_set2([Belief1|Beliefs1],N1,[Belief2|Beliefs2],N4) :- multi_uni_vars(Belief1,Uni1), del_doubles(Uni1,Uni2),
  multi_repl_all_uni(Belief1,Uni2,N1,Belief2), length(Uni2,N2), N3 is N1+N2, uni_belief_set2(Beliefs1,N3,Beliefs2,N4).

% Looking for all the variables in a formula
var_in_formula(var(X),[var(X)]) :- !.
var_in_formula([X|Xs],Ls) :- !, var_in_formula(X,Y), var_in_formula(Xs,Ys), append(Y,Ys,Ls).
var_in_formula(pred(X,Xs),Ls) :- !, var_in_formula(X,Y), var_in_formula(Xs,Ys), append(Y,Ys,Ls).
var_in_formula(_,[]).
% Looking for all the variables in a belief
var_in_belief([],[]).
var_in_belief([X|Xs],Ls) :- var_in_formula(X,Y), var_in_belief(Xs,Ys), append(Y,Ys,Ls).
% renaming the variables
rename_vars(N,[],[],N).
rename_vars(N1,[var(X)|Vars1],[corr(var(X),var(N2))|Vars2],N3) :- N2 is N1+1, rename_vars(N2,Vars1,Vars2,N3).
% upgrading the variables
upgrade_vars(N,Belief,Beliefs,Beliefs,N) :- var_in_belief(Belief,[]), !.
upgrade_vars(N1,Belief,Beliefs1,Beliefs2,N2) :- var_in_belief(Belief,Vars1), del_doubles(Vars1,Vars2),
  rename_vars(N1,Vars2,Vars3,N2), multi_replace(Beliefs1,Vars3,Beliefs2).

% ------------------------------------------------------------------------------------
% Useful predicates for the flexible groups of premises
% ------------------------------------------------------------------------------------

% Checking the correlation between a flexible subformula and an actual corresponding subformula
micro_flex_corr(X,X) :- !.
micro_flex_corr(free(_),_) :- !.
micro_flex_corr(var(_),_) :- !.
micro_flex_corr(_,var(_)) :- !.
micro_flex_corr([X|Xs],[Y|Ys]) :- micro_flex_corr(X,Y), micro_flex_corr(Xs,Ys).
micro_flex_corr(pred(X,Xs),pred(Y,Ys)) :- micro_flex_corr(X,Y), micro_flex_corr(Xs,Ys).
% Checking the correlation between a flexible subformula and all the actual corresponding subformulae
% and returning the number of those corresponding formulae
macro_flex_corr(_,[],0) :- !.
macro_flex_corr(X,[Y|Ys],L1) :- !, micro_flex_corr(X,Y), macro_flex_corr(X,Ys,L), L1 is L+1.
% Checking the correlation between a flexible formula and the actual corresponding formula
form_flex_corr(Name,all(in,Name,X),Xs,L) :- macro_flex_corr(X,Xs,L), !.
form_flex_corr(Name,[X|_],[Y|_],L) :- form_flex_corr(Name,X,Y,L), !.
form_flex_corr(Name,[_|Xs],[_|Ys],L) :- !, form_flex_corr(Name,Xs,Ys,L).
form_flex_corr(Name,pred(X,Xs),pred(X,Ys),L) :- form_flex_corr(Name,Xs,Ys,L).
% Checking the correlation between a flexible belief and the actual corresponding belief
belief_flex_corr(Name,[X|_],[Y|_],L) :- form_flex_corr(Name,X,Y,L), !.
belief_flex_corr(Name,[_|Xs],[_|Ys],L) :- !, belief_flex_corr(Name,Xs,Ys,L).

% Creating new variables replacing a flexible variable
vars4flex(_,0,corr(free(_),[])) :- !.
vars4flex(N,L,corr(free(X),[var(N1)|Vars])) :- N1 is N+1, L1 is L-1, vars4flex(N1,L1,corr(free(X),Vars)).
% Looking for all the unknown flexible variables in a premise
flex_vars(_,N,_,Corrs,[],N,Corrs) :- !.
flex_vars(Name,N1,L,Corrs1,all(_,Name,X),N2,Corrs2) :- !, flex_vars(Name,N1,L,Corrs1,X,N2,Corrs2).
flex_vars(_,N1,L,Corrs,free(Y),N2,[corr(free(Y),Vars)|Corrs]) :- \+ member(corr(free(Y),_),Corrs), !,
  vars4flex(N1,L,corr(free(Y),Vars)), N2 is N1+L.
flex_vars(Name,N1,L,Corrs1,[X|Xs],N3,Corrs3) :- !, flex_vars(Name,N1,L,Corrs1,X,N2,Corrs2),
  flex_vars(Name,N2,L,Corrs2,Xs,N3,Corrs3).
flex_vars(Name,N1,L,Corrs1,pred(_,Xs),N2,Corrs2) :- !, flex_vars(Name,N1,L,Corrs1,Xs,N2,Corrs2).
flex_vars(_,N,_,Corrs,_,N,Corrs).

flex_first([],[],[]).
flex_first([corr(free(X),[Y|Ys])|Corrs1],[corr(free(X),Y)|Corrs2],[corr(free(X),Ys)|Corrs3]) :- flex_first(Corrs1,Corrs2,Corrs3).

% Replacing some flexible variables
repl_flex(Corrs,free(X),var(Y)) :- member(corr(free(X),var(Y)),Corrs), !.
repl_flex(Corrs,[X|Xs],[Y|Ys]) :- !, repl_flex(Corrs,X,Y), repl_flex(Corrs,Xs,Ys).
repl_flex(Corrs,pred(X,Xs),pred(Y,Ys)) :- !, repl_flex(Corrs,X,Y), repl_flex(Corrs,Xs,Ys).
repl_flex(_,X,X).
repl_flex_in(all(out,Name,_),_,Corrs,all(out,Name,Y),Z) :- !, repl_flex(Corrs,Y,Z).
repl_flex_in(all(in,Name,_),0,_,all(in,Name,_),[]) :- !.
repl_flex_in(all(in,Name,X),L,Corrs1,all(in,Name,Y),[Z|Zs]) :- !, flex_first(Corrs1,Corrs2,Corrs3),
  repl_flex(Corrs2,Y,Z), L1 is L-1, repl_flex_in(all(in,Name,X),L1,Corrs3,all(in,Name,Y),Zs).
repl_flex_in(all(S,Name,X),L,Corrs,[Y|Ys],[Z|Zs]) :- !, repl_flex_in(all(S,Name,X),L,Corrs,Y,Z),
  repl_flex_in(all(S,Name,X),L,Corrs,Ys,Zs).
repl_flex_in(all(S,Name,X),L,Corrs,pred(Y,Ys),pred(Z,Zs)) :- !, repl_flex_in(all(S,Name,X),L,Corrs,Y,Z),
  repl_flex_in(all(S,Name,X),L,Corrs,Ys,Zs).
repl_flex_in(_,_,_,X,X).
repl_flex_out(_,0,_,_,[]) :- !.
repl_flex_out(all(_,Name,X),L,Corrs1,Prem,[Y|Ys]) :- flex_first(Corrs1,Corrs2,Corrs3),
  repl_flex_in(all(out,Name,X),1,Corrs2,Prem,Y), L1 is L-1, repl_flex_out(all(_,Name,X),L1,Corrs3,Prem,Ys).

% Modifying the premises
repl_all_flex(all(in,Name,X),N1,L,Corrs1,rule([],Concl1),rule([],Concl2),N2) :- !,
  flex_vars(Name,N1,L,Corrs1,Concl1,N2,Corrs2), repl_flex_in(all(in,Name,X),L,Corrs2,Concl1,Concl2).
repl_all_flex(all(in,Name,X),N1,L,Corrs1,rule([Prem1|Prems1],Concl1),rule(Prems2,Concl2),N3) :-
  is_in(all(in,Name,_),Prem1), !, flex_vars(Name,N1,L,Corrs1,Prem1,N2,Corrs2),
  repl_flex_in(all(in,Name,X),L,Corrs2,Prem1,Prem2),
  repl_all_flex(all(in,Name,X),N2,L,Corrs2,rule([Prem2|Prems1],Concl1),rule(Prems2,Concl2),N3).
repl_all_flex(all(in,Name,X),N1,L,Corrs1,rule([Prem1|Prems1],Concl1),rule(Prems3,Concl2),N3) :-
  is_in(all(out,Name,_),Prem1), !, flex_vars(Name,N1,L,Corrs1,Prem1,N2,Corrs2),
  repl_flex_out(all(in,Name,X),L,Corrs2,Prem1,Prem2),
  repl_all_flex(all(in,Name,X),N2,L,Corrs2,rule(Prems1,Concl1),rule(Prems2,Concl2),N3), append(Prem2,Prems2,Prems3).
repl_all_flex(all(in,Name,X),N1,L,Corrs,rule([Prem1|Prems1],Concl1),rule([Prem1|Prems2],Concl2),N2) :-
  repl_all_flex(all(in,Name,X),N1,L,Corrs,rule(Prems1,Concl1),rule(Prems2,Concl2),N2).

% ------------------------------------------------------------------------------------
% Cleaning the variables (minimum number of variables inside the variables)
% obtained with the correlation function
% ------------------------------------------------------------------------------------

% Checking that there are not 2 or more different interpretations for a same variable
% (the 'replace' predicate is almost only used for resolving the +/-/min/max functions)
clean_vars2([],[]) :- !.
clean_vars2([corr(X,Y)|Vars1],Vars2) :- is_in(X,Y), !, clean_vars2(Vars1,Vars2), nl,nl,nl,print(test),nl,nl,nl.
clean_vars2([corr(X,Y1)|Vars1],[corr(X,Y2)|Vars2]) :- \+ member(corr(X,_),Vars1), \+ is_in_all(X,Vars1),
  \+ is_in_itself(Y1,Vars1), !, replace(Y1,[],Y2), clean_vars2(Vars1,Vars2).
clean_vars2([corr(X,Y1)|Vars1],Vars5) :- select(corr(X,Y2),Vars1,Vars2), !, replace(Y1,[],Y1bis), replace(Y2,[],Y2bis),
  var_reduce(Y1bis,Y2bis,Y3,Vars3), append(Vars2,Vars3,Vars4), clean_vars2([corr(X,Y3)|Vars4],Vars5).
clean_vars2([corr(X1,Y1)|Vars1],Vars3) :- select(corr(X2,Y2),Vars1,Vars2), is_in(X1,Y2), !, replace(Y1,[],Y1bis),
  replace(Y2,[corr(X1,Y1bis)],Y3), replace(Y3,[],Y3bis), clean_vars2([corr(X1,Y1bis),corr(X2,Y3bis)|Vars2],Vars3).
clean_vars2([corr(X,Y1)|Vars1],Vars3) :- replace(Y1,[],Y2), append(Vars1,[corr(X,Y2)],Vars2), clean_vars2(Vars2,Vars3).
clean_vars(X,Z) :- clean_vars2(X,Y), X\=Y, !, clean_vars(Y,Z).
clean_vars(X,X) :- clean_vars2(X,X).

% Reduce the degrees of liberty in a variable when its different definitions
% are not totally the sames but are compatible.
var_reduce(X,X,X,[]) :- !.
var_reduce(var(X),Y,Y,[corr(var(X),Y)]) :- !.
var_reduce(Y,var(X),Y,[corr(var(X),Y)]) :- !.
var_reduce([X|Xs],[Y|Ys],[Z|Zs],Ls3) :- var_reduce(X,Y,Z,Ls1), var_reduce(Xs,Ys,Zs,Ls2), append(Ls1,Ls2,Ls3).
var_reduce(pred(X,Xs),pred(Y,Ys),pred(Z,Zs),Ls3) :- var_reduce(X,Y,Z,Ls1), var_reduce(Xs,Ys,Zs,Ls2), append(Ls1,Ls2,Ls3).

% Checking if a certain element (like a variable) is in a formula
is_in(X,X).
is_in(X,[Y|_]) :- is_in(X,Y).
is_in(X,[_|Ys]) :- is_in(X,Ys).
is_in(X,pred(Y,_)) :- is_in(X,Y).
is_in(X,pred(_,Ys)) :- is_in(X,Ys).
is_in_all(X,Ys) :- member(corr(_,Y),Ys), is_in(X,Y).
is_in_itself(X,Ys) :- member(corr(Y,_),Ys), is_in(Y,X).

% ------------------------------------------------------------------------------------
% Inverting a simple operation (with additions ans substractions)
% ------------------------------------------------------------------------------------
invert_op(X,X,Y,Y) :- !.
invert_op(X,[+|Xs1],Y,Zs) :- select(Ls,Xs1,Xs2), is_in(X,Ls), !, invert_op(X,Ls,[-,Y|Xs2],Zs).
invert_op(X,[-,Ls|Xs],Y,Zs) :- is_in(X,Ls), !, invert_op(X,Ls,[+,Y|Xs],Zs).
invert_op(X,[-,Z|Xs1],Y,Zs) :- select(Ls,Xs1,Xs2), is_in(X,Ls), invert_op(X,Ls,[-,Z,Y|Xs2],Zs).

% ------------------------------------------------------------------------------------
% Conditional predicate
% ------------------------------------------------------------------------------------

condition(r_csf,_,[Alpha,Beta]) :- correlation(Beta,Alpha,Vars), clean_vars(Vars,_).
condition(f_forms,Beliefs,Prem) :- member(Belief,Beliefs), multi_correl(Prem,Belief,Vars), clean_vars(Vars,_).
condition(=,_,[Alpha,Beta]) :- correlation(Alpha,Beta,Vars), clean_vars(Vars,_).
condition(<,_,[X,Y]) :- number(X), number(Y), X<Y.

% Filtering a belief set
belief_filter([],_,[]) :- !.
belief_filter([Belief|Beliefs1],Model,[Belief|Beliefs2]) :- multi_correl(Model,Belief,Vars),
  clean_vars(Vars,_), !, belief_filter(Beliefs1,Model,Beliefs2).
belief_filter([_|Beliefs1],Model,Beliefs2) :- belief_filter(Beliefs1,Model,Beliefs2).

% Finding all the closed subformulae in a belief set.
all_csf([not,pred(X,Xs)],Ls2) :- X \= k, X \= contra, X \= loses, !, all_csf(Xs,Ls1), append([[not,pred(X,Xs)]],Ls1,Ls2).
all_csf(pred(X,Xs),Ls2) :- X \= k, X \= contra, X \= loses, !, all_csf(Xs,Ls1), append([pred(X,Xs)],Ls1,Ls2).
all_csf([X|Xs],Ls) :- !, all_csf(X,Y), all_csf(Xs,Ys), append(Y,Ys,Ls).
all_csf(pred(k,Xs),Ys) :- !, all_csf(Xs,Ys).
all_csf(_,[]).

% ------------------------------------------------------------------------------------
% Cleaning the conclusions
% ------------------------------------------------------------------------------------

% looking for all the universal variables in a formula of a conclusion
uni_vars_back([],[]) :- !.
uni_vars_back(var(X),[var(X)]) :- !.
uni_vars_back([X|Xs],Ls) :- !, uni_vars_back(X,Y), uni_vars_back(Xs,Ys), append(Y,Ys,Ls).
uni_vars_back(pred(X,Xs),Ls) :- !, uni_vars_back(X,Y), uni_vars_back(Xs,Ys), append(Y,Ys,Ls).
uni_vars_back(_,[]).
% looking for all the universal variables in a conclusion
multi_uni_vars_back([],[]).
multi_uni_vars_back([X|Xs],Ls) :- uni_vars_back(X,Y), multi_uni_vars_back(Xs,Ys), append(Y,Ys,Ls).

% setting the invert correlations
any2var(_,[],[]).
any2var(N,[var(X)|Vars],[corr(var(X),any(N1))|Corrs]) :- N1 is N+1, any2var(N1,Vars,Corrs).

% Replacing an entire conclusion
repl_concl(Concl1,Concl2) :- multi_uni_vars_back(Concl1,Vars1), del_doubles(Vars1,Vars2),
  any2var(0,Vars2,Vars3), replace(Concl1,Vars3,Concl2).
% Replacing an entire conclusions set
repl_all_concl([],[]).
repl_all_concl([rule(_,Concl1)|Rules],[Concl2|Concls]) :- repl_concl(Concl1,Concl2), repl_all_concl(Rules,Concls).
